"""Test the driver correctly parses a tags file and responds with correct data."""
from unittest import mock

import pytest

from watlow import command_line
from watlow.mock import Gateway


@pytest.fixture
def gateway_driver(address='fake ip', **kwargs):
    """Confirm the gateway driver correctly initializes."""
    return Gateway(address, **kwargs)


@mock.patch('watlow.Gateway', Gateway)
def test_driver_cli(capsys):
    """Confirm the commandline interface works with reading a zone."""
    command_line(['fakeip', '-z', '1'])
    captured = capsys.readouterr()
    assert 'actual' in captured.out
    assert 'setpoint' in captured.out
    assert 'output' in captured.out


@mock.patch('watlow.Gateway', Gateway)
def test_driver_cli_setpoint(capsys):
    """Confirm the commandline interface works with changing a setpoint."""
    command_line(['fakeip', '-z', '1', '-f', '26'])
    captured = capsys.readouterr()
    assert '"setpoint": 26' in captured.out
    assert '"actual": 26' in captured.out
    with pytest.raises(SystemExit):
        command_line(['fakeip', '-z', 'bogus'])


@pytest.mark.asyncio
async def test_roundtrip(gateway_driver):
    """Confirm various datatypes are read back correctly after being set."""
    expected = {'setpoint': 25.0, 'actual': 25, 'output': 0}
    assert await gateway_driver.get(zone=1) == pytest.approx(expected, rel=1e-6)
    await gateway_driver.set_setpoint(zone=1, setpoint=12.3)
    expected = {'setpoint': 12.3, 'actual': 24, 'output': 0}
    assert await gateway_driver.get(zone=1) == pytest.approx(expected, rel=1e-6)


@pytest.mark.asyncio
async def test_set_errors(gateway_driver):
    """Confirm the driver gives an error on invalid set_setpoint() calls."""
    with pytest.raises(ValueError, match='not in the valid range'):
        await gateway_driver.set_setpoint(zone=1, setpoint=9000)



"""Base functionality for modbus communication.

Distributed under the GNU General Public License v2
Copyright (C) 2022 NuMat Technologies
"""

import asyncio

import pymodbus.exceptions
from pymodbus.client import AsyncModbusTcpClient  # 3.x


class AsyncioModbusClient:
    """A generic asyncio client.

    This expands upon the pymodbus AsyncModbusTcpClient by
    including standard timeouts, async context manager, and queued requests.
    """

    def __init__(self, address: str, timeout: float=1) -> None:
        """Set up communication parameters."""
        self.ip = address
        self.timeout = timeout
        self._detect_pymodbus_version()
        self.client = AsyncModbusTcpClient(address, timeout=timeout)
        self.lock = asyncio.Lock()
        self.connectTask = asyncio.create_task(self._connect())

    async def __aenter__(self):
        """Asynchronously connect with the context manager."""
        return self

    async def __aexit__(self, *args) -> None:
        """Provide exit to the context manager."""
        await self._close()

    def _detect_pymodbus_version(self) -> None:
        major, minor, _patch = map(int, pymodbus.__version__.split('.')[:3])
        self.pymodbus30plus = major == 3
        self.pymodbus32plus = major == 3 and minor >= 2
        self.pymodbus33plus = major == 3 and minor >= 3

    async def _connect(self):
        """Start asynchronous reconnect loop."""
        try:
            await asyncio.wait_for(self.client.connect(), timeout=self.timeout)
        except Exception as e:
            raise OSError(f"Could not connect to '{self.ip}'.") from e

    async def read_coils(self, address: int, count: int):
        """Read modbus output coils (0 address prefix)."""
        return await self._request('read_coils', address, count)

    async def read_registers(self, address: int, count: int) -> list:
        """Read modbus registers.

        The Modbus protocol doesn't allow responses longer than 250 bytes
        (ie. 125 registers, 62 DF addresses), which this function manages by
        chunking larger requests.
        """
        registers = []
        while count > 124:
            r = await self._request('read_holding_registers', address, 124)
            registers += r.registers
            address, count = address + 124, count - 124
        r = await self._request('read_holding_registers', address, count)
        registers += r.registers
        return registers

    async def read_holding_registers(self, address: int, count: int):
        """Read modbus holding registers."""
        await self._request('read_holding_registers', address, count)

    async def write_coil(self, address: int, value):
        """Write a modbus coil."""
        await self._request('write_coil', address, value)

    async def write_coils(self, address: int, values):
        """Write modbus coils."""
        await self._request('write_coils', address, values)

    async def write_register(self, address: int, value, skip_encode=False):
        """Write a modbus register."""
        return await self._request('write_register', address, value,
                                   skip_encode=skip_encode)

    async def write_registers(self, address: int, values, skip_encode=False):
        """Write modbus registers.

        The Modbus protocol doesn't allow requests longer than 250 bytes
        (ie. 125 registers, 62 DF addresses), which this function manages by
        chunking larger requests.
        """
        while len(values) > 62:
            await self._request('write_registers',
                                address, values, skip_encode=skip_encode)
            address, values = address + 124, values[62:]
        await self._request('write_registers',
                            address, values, skip_encode=skip_encode)

    async def _request(self, method, *args, **kwargs):
        """Send a request to the device and awaits a response.

        This mainly ensures that requests are sent serially, as the Modbus
        protocol does not allow simultaneous requests (it'll ignore any
        request sent while it's processing something). The driver handles this
        by assuming there is only one client instance. If other clients
        exist, other logic will have to be added to either prevent or manage
        race conditions.
        """
        await self.connectTask
        async with self.lock:
            try:
                if self.pymodbus32plus:
                    future = getattr(self.client, method)
                else:
                    future = getattr(self.client.protocol, method)  # type: ignore
                return await future(*args, **kwargs)
            except (asyncio.TimeoutError, pymodbus.exceptions.ConnectionException) as e:
                raise TimeoutError("Not connected to Watlow gateway") from e

    async def _close(self) -> None:
        """Close the TCP connection."""
        if self.pymodbus33plus:
            self.client.close()  # 3.3.x
        else:
            await self.client.close()  # type: ignore  # 3.0.x - 3.2.x



"""Mock Watlow interface. Use for debugging systems."""

import struct
from unittest.mock import MagicMock

from pymodbus.datastore import ModbusSparseDataBlock

from watlow.driver import Gateway as realGateway

try:
    from pymodbus.pdu.register_message import ReadHoldingRegistersResponse  # type: ignore
    pymodbus38plus = True
except ImportError:
    pymodbus38plus = False
    try:  # pymodbus 3.7.x
        from pymodbus.pdu.register_read_message import ReadHoldingRegistersResponse  # type: ignore
    except ImportError:
        from pymodbus.register_read_message import ReadHoldingRegistersResponse  # type: ignore

class AsyncClientMock(MagicMock):
    """Magic mock that works with async methods."""

    async def __call__(self, *args, **kwargs):
        """Convert regular mocks into into an async coroutine."""
        return super().__call__(*args, **kwargs)

    async def close(self):
        """Close the connection."""
        ...


class Gateway(realGateway):
    """Mock interface to the Watlow Gateway used to communicate with ovens."""

    def __init__(self, *args, max_temp=220, **kwargs):
        self.setpoint_range = (10, max_temp)
        self.client = AsyncClientMock()
        self._detect_pymodbus_version()
        self.actual_temp_address = 360
        self.setpoint_address = 2160
        self.output_address = 1904
        self.modbus_offset = 5000

        self._registers = ModbusSparseDataBlock({})
        # Initialize each zone with 25.0 for actual/setpoint, 0.0 output
        for zone in range(1, 9):
            zone_offset = self.modbus_offset * (zone - 1)
            for addr, val in [
                (self.actual_temp_address + zone_offset, 25.0),
                (self.setpoint_address + zone_offset, 25.0),
                (self.output_address + zone_offset, 0.0),
            ]:
                hi, lo = struct.unpack('>HH', struct.pack('>f', val))
                self._registers.setValues(addr, [hi, lo])

    def _read_float(self, addr: int) -> float:
        regs = self._registers.getValues(addr, 2)
        assert isinstance(regs, list)
        packed = struct.pack('>HH', regs[0], regs[1])
        return struct.unpack('>f', packed)[0]

    def _write_float(self, addr: int, val: float):
        hi, lo = struct.unpack('>HH', struct.pack('>f', val))
        self._registers.setValues(addr, [hi, lo])

    def _perturb(self):
        for zone in range(1, 9):
            zone_offset = self.modbus_offset * (zone - 1)
            actual_addr = self.actual_temp_address + zone_offset
            setpoint_addr = self.setpoint_address + zone_offset
            output_addr = self.output_address + zone_offset

            actual = self._read_float(actual_addr)
            setpoint = self._read_float(setpoint_addr)
            output = self._read_float(output_addr)

            if actual < setpoint:
                actual += 1
                output = min(setpoint - actual, 100)
            elif actual > setpoint:
                actual -= 1
                output = 0

            self._write_float(actual_addr, actual)
            self._write_float(setpoint_addr, setpoint)
            self._write_float(output_addr, output)

    async def _request(self, method, address, count, **kwargs):
        if method == 'read_holding_registers':
            regs = self._registers.getValues(address, count)
            if pymodbus38plus:
                assert isinstance(regs, list)
                return ReadHoldingRegistersResponse(registers=regs)  # type: ignore
            return ReadHoldingRegistersResponse(regs)  # type: ignore

        if method == 'write_registers':
            for i, val in enumerate(count):
                self._registers.setValues(address + i, val)
            self._perturb()
            return

        raise NotImplementedError(f'Unrecognised method: {method}')
